import re
from urllib.parse import quote

_to_snake_pattern = re.compile(r'(?!^)([A-Z]+)')
_to_camel_pattern = re.compile(r'_([a-z])')


def to_snake(camel_string: str):
    return _to_snake_pattern.sub('_\1', camel_string)


def to_spaced(camel_string: str):
    return to_snake(camel_string).replace('_', ' ')


def to_camel(string: str):
    string = '_'.join(string.split())
    return _to_camel_pattern.sub(lambda match: match[1].upper(), string)


def urlstring(string: str):
    return quote(string.strip())



import typing
from datetime import date, time

import sqlalchemy as sa
from sqlalchemy.sql import ColumnElement

from . import base

SortableTypeDef: typing.TypeAlias = int | float | date | time

SequenceTypeDef: typing.TypeAlias = list | tuple


def equals(attr: ColumnElement, expected: typing.Any) -> base.ComparisonOutput:
    return attr == expected


def not_equals(
    attr: ColumnElement, expected: typing.Any
) -> base.ComparisonOutput:
    return attr != expected


def greater(
    attr: ColumnElement, expected: SortableTypeDef
) -> base.ComparisonOutput:
    return attr > expected


def greater_equals(
    attr: ColumnElement, expected: SortableTypeDef
) -> base.ComparisonOutput:
    return attr >= expected


def lesser(
    attr: ColumnElement, expected: SortableTypeDef
) -> base.ComparisonOutput:
    return attr < expected


def lesser_equals(
    attr: ColumnElement, expected: SortableTypeDef
) -> base.ComparisonOutput:
    return attr <= expected


def between(
    attr: ColumnElement,
    expected: tuple[SortableTypeDef, SortableTypeDef],
) -> base.ComparisonOutput:
    left, right = expected
    return attr.between(left, right)


def like(attr: ColumnElement, expected: str) -> base.ComparisonOutput:
    return attr.like(f'%{expected}%')


def ilike(attr: ColumnElement, expected: str) -> base.ComparisonOutput:
    return attr.ilike(f'%{expected}%')


def contains(
    attr: ColumnElement, expected: SequenceTypeDef
) -> base.ComparisonOutput:
    return attr.in_(expected)


def excludes(
    attr: ColumnElement, expected: SequenceTypeDef
) -> base.ComparisonOutput:
    return attr.not_in(expected)


def isnull(attr: ColumnElement, expected: bool) -> base.ComparisonOutput:
    return attr.is_(None) if expected else attr.is_not(expected)


def always_true(
    attr: ColumnElement, expected: typing.Any
) -> base.ComparisonOutput:
    del attr, expected
    return sa.true()



from http import HTTPStatus
from typing import NamedTuple, final

from .exceptions import _append_exclamation_mark


class _ExtendedResponse(NamedTuple):
    message: str
    status_code: int
    error_key: str


class ExtendedError(Exception):
    def __init__(
        self,
        message: str,
        error_key: str,
        status_code: HTTPStatus = HTTPStatus.BAD_REQUEST,
    ) -> None:
        self._message = message
        self._error_key = error_key
        self._status_code = status_code

    @final
    @property
    def message(self):
        return _append_exclamation_mark(self._message)

    @final
    def response(self) -> _ExtendedResponse:
        return _ExtendedResponse(
            self.message,
            self._status_code,
            self._error_key,
        )



from typing import Generic, TypeVar

import fastapi
import orjson
import pydantic
import pydantic.generics

from souswift_core.utils import string


def _orjson_dumps(val, *, default=None):
    return orjson.dumps(val, default=default).decode()


class Model(pydantic.BaseModel):
    class Config:
        allow_population_by_field_name = True
        json_loads = orjson.loads
        json_dumps = _orjson_dumps
        frozen = True

        @classmethod
        def alias_generator(cls, field: str):
            return string.to_camel(field.removesuffix('_'))


class FrozenModel(Model):
    """Same as Model, just for compatibility"""


T = TypeVar('T', bound=Model)


class Details(FrozenModel):
    last_id: int | None = None
    total_pages: int | None = None
    info: str | None = None
    total_left: int | None = None


class BaseResponse(pydantic.generics.GenericModel, Generic[T]):
    data: list[T]
    details: Details | None = pydantic.Field(default_factory=Details)


class Page(FrozenModel):
    last_id: int = fastapi.Query(0, ge=0)
    per_page: int = fastapi.Query(20, ge=5, le=100)


class BaseError(Model):
    title: str
    path: str
    status: int
    detail: str


class ExtendedBaseError(BaseError):
    error_key: str



from souswift_core.exc import exceptions
from souswift_core.exc.ext import ExtendedError
from souswift_core.exc.handlers import add_exception_handlers

__all__ = ['exceptions', 'add_exception_handlers', 'ExtendedError']



import logging

import sqlalchemy as sa
from context_handler.ext import sqlalchemy
from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import create_async_engine
from starlette.datastructures import State

from souswift_core.providers.config import DatabaseConfig


class DatabaseProvider(sqlalchemy.AsyncSaAdapter):
    state_name = 'database_provider'

    def __init__(
        self, database_config: DatabaseConfig, logger: logging.Logger
    ) -> None:
        self.config = database_config
        self.logger = logger
        engine, self._sync_engine = self._create_engine()
        super().__init__(engine=engine)

    def _create_engine(self):
        return (
            create_async_engine(
                self.config.get_uri(is_async=True), **self.config.pool_config
            ),
            create_engine(
                self.config.get_uri(is_async=False), **self.config.pool_config
            ),
        )

    async def health_check(self):
        async with self.context().begin() as connection:
            await connection.execute(sa.text('SELECT 1'))
        self.logger.info('Database Connection Succeeded')


def setup_database(
    config: DatabaseConfig,
    logger: logging.Logger,
):
    async def _setup_database(state: State):
        provider = DatabaseProvider(config, logger)
        setattr(
            state,
            DatabaseProvider.state_name,
            provider,
        )
        await provider.health_check()

    return _setup_database



import typing

from sqlalchemy.sql import Select

from souswift_core.filters import comp, where


class Paginate(typing.Protocol):
    def __call__(self, query: Select) -> Select:
        ...


def limit_offset_paginate(limit: int, page: int) -> Paginate:
    def paginate(query: Select) -> Select:
        return query.limit(limit).offset(page * limit)

    return paginate


def id_paginate(limit: int, last_id: int) -> Paginate:
    def paginate(query: Select) -> Select:
        return query.where(
            where.Field('pk', last_id, comp=comp.greater).where(
                query.selected_columns
            )
        ).limit(limit)

    return paginate


def _null_paginate(limit: int, offset: int) -> Paginate:
    del limit, offset

    def paginate(query: Select) -> Select:
        return query

    return paginate


null_paginate = _null_paginate(0, 0)



import typing

from sqlalchemy import Boolean, Table
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.elements import BooleanClauseList


class Filter:
    _mark: str = ''

    def where(self, entity: type | Table):
        pass

    def __bool__(self):
        return False

    @property
    def mark(self):
        return self._mark

    def add_mark(self, mark: str):
        self._mark = mark
        return self


ComparisonOutput = typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']


class Comparison(typing.Protocol):
    def __call__(
        self, attr: ColumnElement, expected: typing.Any
    ) -> ComparisonOutput:
        """Receives column or function and returns a sqlalchemy comparison"""
        ...



import typing
from types import NoneType

from pydantic.fields import FieldInfo, ModelField, Undefined

from souswift_core.future import config as future_config
from souswift_core.future import models

NOT_FOUND = object()

ProviderConfigT = typing.TypeVar('ProviderConfigT', bound='ProviderConfig')


class ProviderConfig(models.Model):
    _prefix_: str = ''
    _no_prefix_: typing.Sequence[str] = ()

    class Config(models.Model.Config):
        arbitrary_types_allowed = True
        alias_generator = str.upper


def from_env(
    cls: type[ProviderConfigT],
    __config__: future_config.ConfigLike = future_config.Config(),
    __ignore_none__: bool = True,
    **fields,
) -> ProviderConfigT:
    if __ignore_none__:
        fields = {
            key: value for key, value in fields.items() if value is not None
        }
    config_from_field = _config_getter(cls._prefix_, cls._no_prefix_)
    args = {
        field.name: config_from_field(field, __config__)
        for field in cls.__fields__.values()
        if field.name not in fields
    }
    return cls.parse_obj(args | fields)


def _config_getter(prefix: str, no_prefix: typing.Sequence[str]):
    def _config_from_field(
        field: ModelField, config: future_config.ConfigLike
    ):
        default = _get_default(field.field_info)
        type_, optional = _get_type(field.outer_type_)
        names = _get_env_names(field, prefix, no_prefix)
        if not optional:
            return _get_config(config, names, default, field.type_)
        value = _get_config(config, names, default)
        return type_(value) if value else None

    return _config_from_field


def _get_config(
    config: future_config.ConfigLike,
    names: tuple[str, str],
    default: typing.Any,
    cast: future_config.CastType = lambda a: a,
):
    main_name, fallback = names

    def _skip_key_error(name: str):
        try:
            value = config.get(name, cast=cast)
        except KeyError:
            return NOT_FOUND
        else:
            return value

    value, name_value = _skip_key_error(main_name), _skip_key_error(fallback)
    if value is not NOT_FOUND:
        return value
    if name_value is not NOT_FOUND:
        return name_value
    if default is not future_config.MISSING:
        return default
    return future_config.MissingNameError(main_name)


def _get_env_names(
    field: ModelField, prefix: str, no_prefix: typing.Sequence
) -> tuple[str, typing.Any]:
    def _get_env_name(name: str):
        if prefix and name not in no_prefix:
            name = f'{prefix.removesuffix("_")}_{name}'
        return name.upper()

    return _get_env_name(field.alias), _get_env_name(field.name)


def _get_default(field_info: FieldInfo):
    if field_info.default is not Undefined:
        return field_info.default
    if field_info.default_factory is not None:
        return field_info.default_factory()
    return future_config.MISSING


def _get_type(field_type: type) -> tuple[type, bool]:
    is_union = typing.get_origin(field_type) is typing.Union
    if not is_union:
        return field_type, False
    args = typing.get_args(field_type)
    if len(args) != 2 or args[1] is not NoneType:
        raise NotImplementedError(
            f'from env does not support non optional union: <{field_type!r}>'
        )
    return args[0], True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['souswift_core',
 'souswift_core.exc',
 'souswift_core.filters',
 'souswift_core.future',
 'souswift_core.future.providers',
 'souswift_core.providers',
 'souswift_core.utils']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.41,<2.0.0',
 'aiomysql>=0.1.1,<0.2.0',
 'context-handler>=5.0.0,<6.0.0',
 'fastapi>=0.85.0,<0.86.0',
 'orjson>=3.8.0,<4.0.0',
 'pydantic[email]>=1.10.2,<2.0.0',
 'tzdata>=2022.4,<2023.0']

setup_kwargs = {
    'name': 'souswift-core',
    'version': '3.1.0',
    'description': '',
    'long_description': 'None',
    'author': 'Gustavo Correa',
    'author_email': 'self.gustavocorrea@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import typing
from dataclasses import dataclass
from enum import Enum

import sqlalchemy as sa
from sqlalchemy.sql import ColumnCollection, ColumnElement
from sqlalchemy.sql.elements import BooleanClauseList

from souswift_core.filters import comp as cm
from souswift_core.filters import cp
from souswift_core.filters.base import Comparison, Filter
from souswift_core.providers import Driver

Entity = type | sa.Table | ColumnCollection


def _attr(entity: Entity, field: str) -> ColumnElement:
    try:
        if isinstance(entity, sa.Table):
            result = getattr(typing.cast(sa.Table, entity).c, field, None)
        else:
            result = getattr(entity, field, None) or getattr(
                typing.cast(sa.Table, entity).c, field, None
            )
    except AttributeError:
        raise AttributeError(f'{entity} does not have field {field}') from None
    else:
        if result is None:
            raise NotImplementedError
        return result


@dataclass
class Field(Filter):
    field: str
    value: typing.Any | None
    comp: Comparison = cm.equals
    enum_value: bool = False
    sql_func: typing.Callable[[sa.Column], typing.Any] | None = None

    def similar_to(self, where: Filter) -> typing.TypeGuard[Field]:
        return (
            self.field == where.field
            if isinstance(where, type(self))
            else False
        )

    def __post_init__(self):
        if self.field == 'id':
            self.field = 'id_'
        if isinstance(self.value, bool) and self.comp is cm.isnull:
            self.value = sa.true() if self.value else sa.false()
        if isinstance(self.value, Enum):
            if self.enum_value:
                self.value = self.value and self.value.value
            else:
                self.value = self.value and self.value.name

    def where(self, entity: Entity):
        return self.get_column_clause(entity)  # type: ignore

    def get_column_clause(self, entity: Entity):
        if not self:
            return True
        attr = self.retrieve_attr(entity)
        if self.sql_func:
            attr = self.sql_func(attr)  # type: ignore
        return self.comp(attr, self.value)

    def retrieve_attr(self, entity: Entity):
        return _attr(entity, self.field)

    def __bool__(self):
        return self.value is not None


@dataclass(init=False)
class FilterJoins(Filter):
    operator: type[BooleanClauseList]
    filters: tuple[Filter, ...]

    def __init__(self, *filters: Filter) -> None:
        self.filters = filters

    def where(self, entity: type | sa.Table):
        return self.operator(*(f.where(entity) for f in self.filters))

    def __bool__(self):
        return any(self.filters)


class Or(FilterJoins):
    @property
    def operator(self):
        return sa.or_


class And(FilterJoins):
    @property
    def operator(self):
        return sa.and_


def _entity_from_foreign(attr: ColumnElement) -> type:
    return attr.entity.class_


@dataclass(frozen=True)
class ForeignField(Filter):
    field: str
    comp: cp.RelatedComp

    def where(self, entity: type):
        related_attr = _attr(entity, self.field)
        related_entity = _entity_from_foreign(related_attr)
        return self.comp(related_entity, related_attr)

    def __bool__(self):
        return self.comp.__bool__()


class EmptyFilter(Filter):
    def __bool__(self):
        return True


class FilterIterator(Filter):
    def __init__(self, *filters: Filter):
        self._filter = And(*filters)

    def __bool__(self):
        return bool(self._filter)

    def where(self, entity: type | sa.Table):
        return self._filter.where(entity)

    def fields(self):
        return list(self._fields())

    def _fields(self, _filter: FilterJoins | None = None):
        if not _filter:
            _filter = self._filter
        for item in _filter.filters:
            if not item:
                continue
            if isinstance(item, FilterJoins):
                yield from self._fields(item)
            else:
                yield typing.cast(Field, item).field


class DateFilter:
    def __init__(self, driver: Driver):
        self._driver = driver

    @staticmethod
    def format(date_part: int):
        return str(date_part).zfill(2)

    def day(self, column: sa.Column):
        if self._driver is Driver.SQLITE:
            return sa.func.strftime('%d', column)
        return sa.extract('DAY', column)

    def month(self, column: sa.Column):
        if self._driver is Driver.SQLITE:
            return sa.func.strftime('%m', column)
        return sa.extract('MONTH', column)


class JSONFilter:
    def __init__(self, driver: Driver):
        self._driver = driver

    @property
    def contains(self):
        return (
            cp.AlwaysTrue()
            if self._driver is Driver.SQLITE
            else cp.JSONContains()
        )

    @property
    def empty(self):
        return (
            cp.AlwaysTrue()
            if self._driver is Driver.SQLITE
            else cp.EmptyJson()
        )
